 /*

  copyright (C) 1999 _TAG_ (Togna Alessandro Giorgio)

  This was a part of the LiteStep Shell Source code.
  Copyright (C) 1997-98 The LiteStep Development Team

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include <windows.h>
#include <stdio.h>
#include <shlobj.h>
#include <tchar.h>
#include <crtdbg.h>
#include <process.h>
#include "popup.h"

const char rcsRevision[] = "$Revision: 0.2 $"; // Our Version 
const char rcsId[] = "$Id: popup.cpp,v 0.2 1999/03/23 01:09:25 _TAG_ Exp $"; // The Full RCS ID.
char popupFontFace[256];
const char szAppName[] = "LPopupMenu"; // Our window class, etc
static char popupSig[4] = {'P', 'O', 'P', 'S'};

// timer data and timers
static timer submenuTimer(MENU_TIMER);
static timer scrollTimer(SCROLL_TIMER);

const cxGap = 4; // min. gap between icon and text
const cxArrow = 7; // width of folder arrow icon
const cyArrow = 7; // height of folder arrow icon

int popupSubHeight = 20;
int popupMinWidth = 100;
int popupIconSize = 20;
int titleHeight = 20;
int popupFixedWidth = 0;
int popupMenuDelay = 0;
int popupFontHeight;
int popupTextOffset;
int popupTop;

HFONT popupFont;
HICON popupDefaultIcon;

Popup popup; //popup class struct

HINSTANCE appInstance;
HWND itemCaller = NULL;
HWND parentWin;
HWND hMainWnd = NULL;
HPEN hpenBlack = NULL;
HPEN hpenWhite = NULL;
HPEN hpenGray = NULL;
HBRUSH hbrBlack = NULL;
HBRUSH hbrGray = NULL;

BOOL noPopupBevel;
BOOL noPopupPix = FALSE;
BOOL transBlt=TRUE;
BOOL popupFolderIcon=TRUE;
BOOL showPopupIcons = FALSE;
BOOL isFixedWidth = FALSE;
BOOL shrinkPopupBar = FALSE;

BitMP backBMP;
BitMP titleBMP;
BitMP selectBMP;
BitMP bottomBMP;

////////////////////////////////////////////////////////////////////////////////
//BEGIN CLASS DEFINITION 
////////////////////////////////////////////////////////////////////////////////
// ** BitMP Class
BitMP::BitMP() {
	bitmap = NULL;
	name[0] = '\0';
	x = 0;
	y = 0;
	backColor = RGB(255,255,255);
	foreColor = RGB(255,255,255);
}

void BitMP::destroy() {
	if (bitmap) {
		::DeleteObject(bitmap);
		bitmap = NULL;
	}
	x = 0;
	y = 0;
	name[0] = '\0';
	backColor = RGB(255,255,255);
	foreColor = RGB(255,255,255);
}

// ** menuItem Class
menuItem::menuItem() {
	init();
}

menuItem::menuItem(LPSTR lpszText, LPSTR lpszCommand, LPSTR lpszParams, Popup *pop) {
	init();
	strcpy(name, lpszText);
	strcpy(command, lpszCommand);
	if (lpszParams != NULL)
		strcpy(params, lpszParams);
	sub = pop;
}

menuItem::~menuItem() {
	::DeleteObject(hIcon);
}

void menuItem::init() {
	owner = NULL;
	type = mi_basic;
	ZeroMemory(name, sizeof(name));
	ZeroMemory(command, sizeof(command));
	ZeroMemory(params, sizeof(params));
	hIcon = NULL;
	sub = NULL;
}

// ** list Class
template<class T> list<T>::list(int alloc, int inc) {
	allocated = 0;
	count = 0;
	increment = inc;
	data = NULL;
	reallocate(alloc);
}

template<class T> list<T>::~list() {
	if (data)
		delete [] data;
}

template<class T> void list<T>::reallocate(int num) {
	if (num <= 0 && count < allocated )
		return;
	T *pOldList = data;
	int inc = (num > 0) ? num : increment;
	_ASSERT(inc > 0);
	data = new T[allocated + inc];
	allocated += inc;
	// copy items from old list
	// since 'menuItem' class doesn't have any virtual functions,
	// we will use a direct memory copy
	memcpy(data, pOldList, sizeof(T)*count);
	//for ( int i = 0; i < m_used; i++ )
	//   *(m_list[i]) = *(pOldList[i]);
}

template<class T> T& list<T>::operator [] (int i) {
	_ASSERT(data);
	_ASSERT(i >= 0);
	_ASSERT(i < count);
	return *(data+i);
}

template<class T> const T& list<T>::operator [] (int i) const {
	_ASSERT(data);
	_ASSERT(i >= 0);
	_ASSERT(i < count);
	return const_cast<const T&>(*(data+i));
}

template<class T> int list<T>::add(const T& t) {
	reallocate(); // allocate only if not enough room
	_ASSERT(data);
	_ASSERT(count >= 0);
	*(data+count) = t;
	count++;
	return(count-1);
}

template<class T> void list<T>::copy(const list<T>& l) {
	int tocopy = l.count;
	reallocate(tocopy);
	memcpy(data+count, l.data, tocopy*sizeof(T));
	count += tocopy;
}

template<class T> void list<T>::clear() {
	delete [] data;
	allocated = count = 0;
}

template<class T> void list<T>::sort(int(__cdecl *compare )(const void *elem1, const void *elem2 )) {
	qsort(data, count, sizeof(T), compare);
}

template<class T> bool list<T>::validIndex(int indx) const {
	if (indx < 0 || indx >= count)
		return false;
	return true;
}

// ** itemList Class
int itemList::add(LPSTR lpszText, LPSTR lpszCommand, LPSTR lpszParams, Popup *lp) {
 menuItem mi(lpszText, lpszCommand, lpszParams, lp);
 return list<menuItem>::add(mi);
}

// ** timer structure
void timer::setTimer(Popup *p, int indx, int interval) {
	killTimer();
	tmPop = p;
	tmNdx  = indx;
	if (tmPop && tmPop->hwnd) {
		::SetTimer(tmPop->hwnd, id, interval, NULL);
		timerSet = true;
	}
}

void timer::killTimer() {
	if (timerSet && tmPop && tmPop->hwnd)
		::KillTimer(tmPop->hwnd, id);
	timerSet = false;
}

bool timer::sameItem(Popup *p, int indx) const {
	if (tmPop == p && tmNdx == indx)
		return true;
	else
		return false;
}

// ** Popup Class
Popup::Popup(Popup *parentMenu) {
	hwnd = NULL;
	ZeroMemory(title, sizeof(title));
	current = -1;
	first = last = -1;
}

Popup::~Popup() {
}

menuItem* Popup::safeItem(int indx) const {
	if (!this || !list.validIndex(indx))
		return NULL;
	return (menuItem*)&list[indx];
}

int Popup::firstVisible() {
	if (first < 0)
		first = 0;
	if (list.count > 0 && first >= list.count)
		first = list.count - 1;
	return first;
}

int Popup::lastVisible() {
	if (last < 0 || (list.count > 0 && last >= list.count))
		last = list.count - 1;
	return last;
}

void Popup::firstVisible(int n) {
	if ((first = n) < 0)
		first = 0;
}

void Popup::lastVisible(int n) { 
	if ((last = n) > list.count - 1)
		last = list.count - 1;
}

void Popup::scroll(bool down) {
	bool scrolled = false;
	if (down && lastVisible() < list.count - 1) {
		first++, last++;
		scrolled = true;
	} else if (firstVisible() > 0) {
		first--, last--;
		scrolled = true;
	}
	if (scrolled) {
	::InvalidateRect(this->hwnd, NULL, FALSE);
	::UpdateWindow(this->hwnd);
	}
}

bool Popup::createWindow() {
	if ( hwnd )
		return true; // window already exists

	SIZE s;
	numVisible = calcSize(s);

	HWND hParent = parentWin;
	_ASSERT(hParent);
	hwnd = ::CreateWindowEx(
		WS_EX_TOPMOST|WS_EX_TOOLWINDOW, // exstyles
		szAppName,						// our window class name
		"",								// use description for a window title
		WS_POPUP,						// window styles
		0, 0,							// position
		s.cx, s.cy,						// width & height of window
		hParent,						// parent window
		NULL,							// no menu
		appInstance,					// hInstance of DLL
		NULL);							// no window creation data
	if (hwnd) {
		::SetWindowLong(hwnd, GWL_USERDATA, magicDWord);
		LONG l;
		strncpy((char*)&l, popupSig, 4);
		::SetWindowLong(hwnd, 0, l);
		::SetWindowLong(hwnd, 4, (LONG)this);
		// set first/last visibles
		firstVisible(0);
		lastVisible(numVisible-1);
	} else
		_RPT1(_CRT_ERROR, "Failed to create POPUP window; error '%d'\n", ::GetLastError());
	return (bool)(hwnd != NULL);
}

void Popup::destroyWindow() {
	::ShowWindow(hwnd, SW_HIDE);
	::SendMessage(parentWin, LM_POPUPHIDDEN, NULL, NULL);
	current = -1;
}

void Popup::display(int x, int y) {
	SIZE s;
	numVisible = calcSize(s);
	firstVisible(0);
	lastVisible(numVisible-1);
	::SetWindowPos(hwnd, HWND_TOP, x, y, s.cx, s.cy, SWP_SHOWWINDOW);
	::InvalidateRect(hwnd, NULL, FALSE);
	::UpdateWindow(hwnd);
}

void Popup::paint() {
	HDC hdc, memDC;
	PAINTSTRUCT ps;
	HFONT hfntOld;
	RECT rcWnd, rc, rcTmp;
	POINT poly[3];
	LPSTR name;

	name = this->title;
	hdc = ::BeginPaint(hwnd, &ps);
	memDC = ::CreateCompatibleDC(hdc);
	hfntOld = (HFONT)::SelectObject(hdc, popupFont);
	::SetBkMode(hdc, TRANSPARENT);
	::GetClientRect(hwnd, &rcWnd);
	if (::IsRectEmpty(&ps.rcPaint))
		::SetRect(&ps.rcPaint, rcWnd.left, rcWnd.top, rcWnd.right, rcWnd.bottom);
	::SetRect(&rc, 0, 0, rcWnd.right, 0);

	// draw the title
	rc.bottom = titleHeight;
	if (::IntersectRect(&rcTmp, &ps.rcPaint, &rc)) {
		if (!transBlt) {
			::SetBkColor(hdc, titleBMP.backColor );
			::ExtTextOut(hdc, rc.left, rc.top, ETO_OPAQUE, &rc, NULL, 0, NULL);
		}
		paintBMP(hdc, memDC, titleBMP, rc);
		rc.left = popupTextOffset;
		::SetTextColor(hdc, titleBMP.foreColor);
		::DrawText(hdc, name, _tcslen(name), &rc,
			DT_END_ELLIPSIS | DT_VCENTER | DT_SINGLELINE | DT_NOPREFIX);
		rc.left = 0;
	}

	// draw individual menu items
	for (int indx = firstVisible(); indx <= first+numVisible-1; indx++) {
		const menuItem& mi = list[indx];
		rc.top    = rc.bottom;
		rc.bottom = rc.top + popupSubHeight;
		// check if need to repaint
		if (!::IntersectRect(&rcTmp, &ps.rcPaint, &rc))
			continue;
		// determine colors and bitmap to use
		if (indx == current)
			drawItem(mi, hdc, memDC, selectBMP, rc);
		else
			drawItem(mi, hdc, memDC, backBMP, rc);
	}

	// draw the bottom bitmap
	if (bottomBMP.bitmap) {
		::GetClientRect(hwnd, &rc);
		rc.top  = rc.bottom - titleHeight;
		if (::IntersectRect(&rcTmp, &ps.rcPaint, &rc)) {
			::SetBkColor(hdc, bottomBMP.backColor);
			::ExtTextOut(hdc, rc.left, rc.top, ETO_OPAQUE, &rc, NULL, 0, NULL);
			paintBMP(hdc, memDC, bottomBMP, rc);
		}
	}

	// paint scrolling buttons
	::SelectObject(hdc, hpenBlack);
	if (firstVisible() > 0) {
		itemRect(firstVisible(), &rc);
		HBRUSH hbrOld = (HBRUSH)::SelectObject(hdc, hbrGray);
		::Rectangle(hdc, rc.left, rc.top, rc.right, rc.bottom);
		int xmid = rc.left + (rc.right-rc.left)/2;
		int ymid = rc.top + (rc.bottom-rc.top)/2;
		poly[0].x = xmid-4,
		poly[0].y = ymid+5,
		poly[1].x = xmid,
		poly[1].y = ymid-5,
		poly[2].x = xmid+4,
		poly[2].y = ymid+5;
		::SelectObject(hdc, hbrBlack);
		::Polygon(hdc, poly, 3);
		::SelectObject(hdc, hbrOld);
	}
	if (lastVisible() < list.count-1) {
		itemRect(lastVisible(), &rc);
		HBRUSH hbrOld = (HBRUSH)::SelectObject(hdc, hbrGray );
		::Rectangle(hdc, rc.left, rc.top, rc.right, rc.bottom);
		int xmid = rc.left + (rc.right-rc.left)/2;
		int ymid = rc.top + (rc.bottom-rc.top)/2;
		poly[0].x = xmid-4,
		poly[0].y = ymid-5,
		poly[1].x = xmid,
		poly[1].y = ymid+5,
		poly[2].x = xmid+4,
		poly[2].y = ymid-5;
		::SelectObject(hdc, hbrBlack);
		::Polygon(hdc, poly, 3);
		::SelectObject(hdc,  hbrOld);
	}

	// release resources
	::SelectObject(hdc, hfntOld);
	::DeleteDC(memDC);
	::ReleaseDC(hwnd, hdc);
	::EndPaint(hwnd, &ps);
}

void Popup::paintBMP(HDC hDC, HDC memDC, BitMP bmp, RECT r) {
	if (!bmp.bitmap)
		return;

	int h = r.bottom - r.top;
	int w = r.right - r.left;

	SelectObject(memDC, bmp.bitmap);
	if (shrinkPopupBar) {
		StretchBlt(hDC, r.left, r.top, w, h, memDC, 0, 0, bmp.x, bmp.y, SRCCOPY);
	} else {
		int bW, bH;
		h = r.bottom;
		w = r.right;

		for(int x=r.left; x<w; x+=bmp.x)
			for(int y=r.top; y<h; y+=bmp.y) {
				bW = (x+bmp.x) <= w ? bmp.x : x < w ? w-x : w;
				bH = (y+bmp.y) <= h ? bmp.y : y < h ? h-y : h;
				BitBlt(hDC, x, y, bW, bH, memDC, 0, 0, SRCCOPY);
			}
	}
}

void Popup::drawItem(const menuItem& mi, HDC hdc, HDC mem, BitMP bmp, RECT rc) {
	COLORREF fg = bmp.foreColor;
	COLORREF bg = bmp.backColor;
	HBITMAP hbmp = bmp.bitmap;

	if (!transBlt) {
		::SetBkColor(hdc, bg);
		::ExtTextOut(hdc, rc.left, rc.top, ETO_OPAQUE, &rc, NULL, 0, NULL);
	}

	paintBMP(hdc, mem, bmp, rc);
	if (showPopupIcons) {
		HICON hicon;
		hicon = mi.hIcon;
		if (hicon)
			::DrawIconEx(hdc, rc.left+cxGap, (rc.top + rc.bottom - popupIconSize)/2,
				hicon, popupIconSize, popupIconSize,
				0, NULL, DI_NORMAL);
	}
	int offset = max(popupIconSize + 2*cxGap, popupTextOffset );
	rc.left += offset;
	::SetTextColor(hdc, fg);
	::DrawText(hdc, mi.name, _tcslen(mi.name), &rc,
		DT_END_ELLIPSIS | DT_VCENTER | DT_SINGLELINE);
	rc.left -= offset;

	// draw bevel
	if (!noPopupBevel) {    
		HPEN hpenOld = (HPEN)::SelectObject(hdc, hpenWhite);
		::MoveToEx(hdc, 0, rc.bottom-2, NULL);
		::LineTo(hdc, 0, rc.top);
		::LineTo(hdc, rc.right, rc.top);
		::SelectObject(hdc, hpenBlack);
		::MoveToEx(hdc, 0, rc.bottom-1, NULL);
		::LineTo(hdc, rc.right-1, rc.bottom-1);
		::LineTo(hdc, rc.right-1, rc.top-1);
		::SelectObject(hdc, hpenOld);
	}
}

bool Popup::itemRect(int indx, RECT *lpRect) const {
	if (!lpRect)
		return false;
	::SetRectEmpty(lpRect);
	if (!list.validIndex(indx))
		return false;
	if (indx < first || indx > last)
		return false;
	::GetClientRect(hwnd, lpRect);
	// start right below title bitmap
	lpRect->top = titleHeight;
	// skip all preceding menu items
	for (int i = first; i < indx; i++)
		lpRect->top += popupSubHeight;
	// set bottom value
	lpRect->bottom = lpRect->top + popupSubHeight;
	return true;
}

int Popup::calcSize(SIZE& s) const {
	int i, h;
	int n = 0; //number of items to return
	int hmax = ::MulDiv(::GetSystemMetrics( SM_CYSCREEN ), 8, 10);
	int wmax = ::MulDiv(::GetSystemMetrics( SM_CXSCREEN ), 6, 10);
	HDC hdc;
	SIZE ext;

	s.cx = 0,  // accumulated width
	s.cy = 0;  // accumulated height
	if ( 0 >= (h = titleHeight))
		h = popupSubHeight;
	s.cy += h;
	s.cy += titleHeight;
	hdc = ::CreateDC("DISPLAY", NULL, NULL, NULL);
	::SelectObject(hdc, popupFont);
	for (i = 0; i < list.count; i++ ) {
		const menuItem& mi = list[i];
		if (::GetTextExtentPoint32(hdc, mi.name, _tcslen(mi.name), &ext))
			s.cx = max(s.cx, ext.cx);
		if (n == 0) {
			h = popupSubHeight;
			if (s.cy + h > hmax)
				n = i; // number of items to fit into the height
			else
				s.cy += h;
		}
	}
	if (n == 0)
		n = i;
	::DeleteDC(hdc);
	s.cx += max(popupIconSize + 2*cxGap, popupTextOffset);
	s.cx += 2*cxGap + cxArrow;
	if (s.cx > wmax)
		s.cx = wmax;
	if (s.cx < popupMinWidth )
		s.cx = popupMinWidth;
	return n;
}

int Popup::itemAt(POINT pt) const {
	if (!hwnd)
		return -1;
	RECT rc;
	::GetClientRect(hwnd, &rc);
	if (!::PtInRect(&rc, pt))
		return -1;
	// check for title hit
	int h = 0; // total (accumulated) height
	int h2;    // height of individual item
	h += titleHeight;
	if (pt.y < h)
		return -1;
	// scan all items
	for (int i = first; i <= last; i++) {
		h2 = popupSubHeight;
		if (pt.y < h + h2 + 1)
			return i;
		h += h2;
	}
	return -1;
}

bool Popup::highlight(int indx, bool snapCursor) {
	// if indx is out of scope we'll only remove highlight from
	// currently selected item
	RECT rc;
	int old;
	bool valid = true;

	old = current;
	current = indx;
	// if current item out of view, scroll to bring it
	// to view
	if (current >= 0 && current < firstVisible()) {
		firstVisible(current - 1);
		lastVisible(firstVisible() + (numVisible-1));
		::InvalidateRect(hwnd, NULL, FALSE);
	} else
		if (current > lastVisible() && current <= list.count-1) {
			lastVisible(current+1);
			firstVisible(lastVisible() - (numVisible-1));
			::InvalidateRect(hwnd, NULL, FALSE);
		} else {
			if (itemRect(old, &rc))
				::InvalidateRect(hwnd, &rc, FALSE);
			if (true == (valid = itemRect(current, &rc )))
				::InvalidateRect(hwnd, &rc, FALSE );
		}
	if (snapCursor && valid) {
		itemRect(current, &rc);
		rc.top  += 2,
		rc.left += 2;
		::ClientToScreen(hwnd, (POINT*)&rc);
		::SetCursorPos(rc.left, rc.top);
	}
	::UpdateWindow(hwnd);
	return true;
}

int Popup::hittest(POINT pt, int *pindx) const {
	RECT rc;
	::GetClientRect(hwnd, &rc);
	if (!::PtInRect(&rc, pt))
		return HTNOWHERE;
	int indx = -1;
	int h = 0;
	int h2;

	h += titleHeight;
	do {
		if (pt.y < h)
			break;
		if (++indx >= list.count)
			break;
		h2 = popupSubHeight;
		h += h2;
	} while(true);
	if (indx < 0) {
		if (pindx) 
			*pindx = -1;
		return HTCAPTION;
	}
	if (indx >= list.count) {
		if (pindx) 
			*pindx = -1;
		return HTCAPTION;
	}
	if (pindx) 
		*pindx = indx;
	return HTCLIENT;
}

bool Popup::execute(int indx) {
	if (indx < 0)
		indx = current;
	if (!list.validIndex(indx))
		return false;

	const menuItem& mi = list[indx];
	windowType *winList;
	HWND hwnd;
	char title[MAX_PATH];
	int maxWin, n;
	winList = (windowType *)::SendMessage(parentWin, LM_WINLIST, 0, 0);
	maxWin = (int)::SendMessage(parentWin, LM_WINLIST, 1 ,0);
	for (n=0; n<maxWin && winList[n].Handle; n++) {
		hwnd = winList[n].Handle;
		::GetWindowText(hwnd, (char *)&title, MAX_PATH);
		if (!strcmp(mi.name, (char *)&title)) {
			if (::IsIconic(hwnd)) ::OpenIcon(hwnd);
			::SetForegroundWindow(hwnd);
			break;
		}
	}
	return true;
}

void Popup::loadValues() {
	char iname[MAX_TEXT];

//    GetRCString("HotListName", popupMenuName, "", 256);
    GetRCString("HotListName", popup.title, "", 256);
    GetRCString("PopupTitlePix", titleBMP.name	, "", 256);
    GetRCString("PopupEntryPix", backBMP.name	, "", 256);
    GetRCString("PopupSelEntryPix", selectBMP.name	, "", 256);
	GetRCString("PopupBottomPix", bottomBMP.name, "", 256);
    GetRCString("PopupFontFace", popupFontFace, "Arial", 256);
	if (strlen(titleBMP.name)==0 || strlen(backBMP.name)==0 || strlen(selectBMP.name)==0) 
		noPopupPix = TRUE;
	popupFolderIcon = GetRCBool("NoPopupFolderIcon", FALSE);
	transBlt = GetRCBool("NoPopupTransparent", FALSE);
    noPopupBevel = GetRCBool("NoPopupBevel", TRUE);
    popupFontHeight = GetRCInt("PopupFontHeight", 16);
    titleBMP.foreColor = GetRCColor("PopupTitleColor", 0xffffff);
    bottomBMP.foreColor = GetRCColor("PopupTitleColor", 0xffffff);
    backBMP.foreColor = GetRCColor("PopupEntryColor", 0x000000);
    selectBMP.foreColor = GetRCColor("PopupSelEntryColor", 0x000000);
	popupMenuDelay = GetRCInt("PopupMenuDelay", 10);
    popupMinWidth = GetRCInt("minpopupwidth", 100);
    popupSubHeight = GetRCInt("PopupSubmenuHeight", 20);
	titleHeight = GetRCInt("PopupSubmenuHeight", 20);
	popupFixedWidth = GetRCInt("PopupFixedWidth", 0);
	isFixedWidth = popupFixedWidth==0?FALSE:TRUE;
	popupTextOffset = GetRCInt("PopupTextOffset", 5);
	showPopupIcons = GetRCBool("ShowPopupIcons", TRUE);
	if (showPopupIcons) {
		popupIconSize = GetRCInt("PopupIconSize", 20);
		GetRCString("PopupDefaultIcon", iname, "default.ico", 256);
		popupDefaultIcon = LoadLSIcon(iname, NULL);
		popupIconSize = popupIconSize>popupSubHeight?popupSubHeight:popupIconSize;
		popupTop = (popupSubHeight-popupIconSize)/2;
	} else 
		popupIconSize = 0;
	// Bottom bitmap is loaded as there is no setting to indicate whether it is drawn or not - yet ...
	bottomBMP.bitmap = LoadLSImage(bottomBMP.name, NULL);
	if (!noPopupPix) {
	    //loads up images i will eventualy seperate this into its own function
		backBMP.bitmap = LoadLSImage(backBMP.name, NULL);
	    titleBMP.bitmap = LoadLSImage(titleBMP.name, NULL);
		selectBMP.bitmap = LoadLSImage(selectBMP.name, NULL); 

	    GetLSBitmapSize(selectBMP.bitmap, &selectBMP.x, &selectBMP.y);
		GetLSBitmapSize(backBMP.bitmap, &backBMP.x, &backBMP.y);
	    GetLSBitmapSize(titleBMP.bitmap, &titleBMP.x, &titleBMP.y);
		GetLSBitmapSize(bottomBMP.bitmap, &bottomBMP.x, &bottomBMP.y);

		//mian's transparency stuff
		selectBMP.region = BitmapToRegion(selectBMP.bitmap, RGB(255,0,255), 0x10101010, 0, 0);
		titleBMP.region = BitmapToRegion(titleBMP.bitmap, RGB(255,0,255), 0x10101010, 0, 0);
		backBMP.region = BitmapToRegion(backBMP.bitmap, RGB(255,0,255), 0x10101010, 0, 0);
		bottomBMP.region = BitmapToRegion(bottomBMP.bitmap, RGB(255,0,255), 0x10101010, 0, 0); 
		//Bitmap shrink/tile option
		shrinkPopupBar = GetRCBool("ShrinkPopupBar", TRUE);
	}
	titleBMP.backColor = GetRCColor("PopupTitleBack", 0x000000);
	bottomBMP.backColor = GetRCColor("PopupTitleBack", 0x000000);
	backBMP.backColor = GetRCColor("PopupNormalBack", 0xa4a0a0);
	selectBMP.backColor = GetRCColor("PopupSelBack", 0xffffff);
	// Stock objects
	hpenBlack = (HPEN)GetStockObject(BLACK_PEN);
	hpenWhite = (HPEN)GetStockObject(WHITE_PEN);
	hpenGray = (HPEN)CreatePen(PS_SOLID, 1, RGB(0x40,0x40,0x40));
	hbrBlack = (HBRUSH)GetStockObject(BLACK_BRUSH);
	hbrGray = (HBRUSH)GetStockObject(GRAY_BRUSH);
    popupFont = CreateFont(popupFontHeight,
    		0, 0, 0, FW_NORMAL,
    		FALSE, FALSE, FALSE,
    		DEFAULT_CHARSET,
    		OUT_DEFAULT_PRECIS,
    		CLIP_DEFAULT_PRECIS,
    		DEFAULT_QUALITY,
    		DEFAULT_PITCH,
    		popupFontFace);
}

void Popup::buildTaskList() {
	windowType *winList;
	char title[MAX_PATH];
	BOOL noList = TRUE;
	int n, maxWin;
	Popup *pop = &popup;
	int indx;
	menuItem *pItem;

	winList = (windowType *)SendMessage(parentWin, LM_WINLIST, 0, 0);
	maxWin = (int) SendMessage(parentWin, LM_WINLIST, 1, 0);
	for (n = 0; n < maxWin && winList[n].Handle; n++)
	{
		HICON hicon = NULL;
		noList = FALSE;

		SendMessageTimeout(winList[n].Handle, WM_GETICON, 1, 0, 0, 250, (unsigned long *)&hicon);
		if (!hicon) 
			SendMessageTimeout(winList[n].Handle, WM_QUERYDRAGICON, 0, 0, 0, 1000, (unsigned long *)&hicon);
		if (!hicon) 
			hicon = (HICON)GetClassLong(winList[n].Handle, GCL_HICON);
		if (!hicon) 
			hicon = (HICON)GetClassLong(winList[n].Handle, GCL_HICONSM);
		title[0] = 0;
		if (winList[n].Handle != NULL) {
			GetWindowText(winList[n].Handle, (char *)&title, MAX_PATH);
		}
		indx = pop->list.add(title,"T@sk",NULL);
		pItem = pop->safeItem(indx);
		_ASSERT(pItem);
		pItem->owner=pop;
		pItem->type=menuItem::mi_task;
		pItem->hIcon=hicon;
	}
	if (noList) {
		indx = pop->list.add("No Tasks","T@sk",NULL);
		pItem = pop->safeItem(indx);
		_ASSERT(pItem);
		pItem->owner=pop;
		pItem->type=menuItem::mi_task;
	}
}

////////////////////////////////////////////////////////////////////////////////
//END CLASS DEFINITION 
////////////////////////////////////////////////////////////////////////////////
// window procedure for our window
LRESULT CALLBACK wndProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam) {
	Popup *pop = (Popup*)::GetWindowLong(hwnd, 4);

	if (!pop)
		return ::DefWindowProc(hwnd, message, wParam, lParam);

	const menuItem *pItem;
	int indx;
	POINT pt, pts;
	LONG l;
	bool onScroll = false;

	l = ::GetMessagePos();
	pt.x = pts.x = LOWORD(l);
	pt.y = pts.y = HIWORD(l);
	::ScreenToClient(pop->hwnd, &pt);
	indx = pop->itemAt(pt);
	pItem = pop->safeItem(indx);
	if ((indx == pop->firstVisible() && indx > 0) ||
	(indx == pop->lastVisible()  && indx < pop->list.count-1))
		onScroll = true;

	switch (message) {
	case LM_GETREVID: 
	{
		char *buf = (char *) lParam;

		if (wParam == 0) {
			strcpy(buf, "popup.dll: ");
			strcat(buf, &rcsRevision[11]);
			buf[strlen(buf)-1] = '\0';
		} else if (wParam == 1) {
			strcpy(buf, &rcsId[1]);
			buf[strlen(buf)-1] = '\0';
		} else {
			strcpy(buf, "");
		}
		return strlen(buf);
	}

	case LM_LPOPUP :
		pop->list.clear();
		pop->buildTaskList();
		pop->display(pts);
		::SetForegroundWindow(pop->hwnd);
		return 0;
/* _TAG_ */
	case LM_HIDEPOPUP :
		popup.destroyWindow();
		break;
/* END _TAG_ */
	case WM_PAINT :
		pop->paint();
		return 0;

	case WM_ERASEBKGND :
		return 0;

	case WM_LBUTTONDOWN :
		// scrolling support: set timer
		// have to do it before "execute", 'cause
		// first and last visible may change
		if (onScroll)
			scrollTimer.setTimer(pop, indx, SCROLL_DELAY);
		break;

	case WM_LBUTTONUP :
		scrollTimer.killTimer();
		if (!onScroll)
			pop->execute(indx);
		break;

	case WM_MOUSEMOVE :
	{
		if (pop->current != indx)
			pop->highlight(indx, false);
		bool other = !submenuTimer.sameItem(pop, indx);
		if (!pItem || other ) {
			submenuTimer.killTimer();
		}
		if (other) submenuTimer.resetIndex();
		break;
	}

	case WM_TIMER :
		switch (wParam) {
		case MENU_TIMER :
		{
			submenuTimer.killTimer();
			pop->execute(indx);
			break;
		}
		case SCROLL_TIMER :
			pop->scroll((bool)(indx == pop->lastVisible()));
			break;
		}
		break;

	case WM_NCHITTEST :
	{
		int ht = pop->hittest(pt);
		return ht;
		break;
	}

	case WM_ACTIVATE :
		if (LOWORD(wParam) == WA_INACTIVE) {
			HWND hwndGetFocus = (HWND)lParam;
			submenuTimer.killTimer();
			popup.destroyWindow();
			return 0;
		}
		break;

	case WM_SYSCOMMAND :
		if (wParam == SC_CLOSE) {
			::PostMessage(parentWin, WM_KEYDOWN, LM_SHUTDOWN, 0);
			return 0;
		}
		break;

	case WM_KEYDOWN :
	switch (wParam) {
	case VK_DOWN :
		pop->highlight(pop->current+1, true);
		break;

	case VK_UP :
		pop->highlight(pop->current-1, true);
		break;

	case VK_HOME :
		pop->highlight(0, true);
		break;

	case VK_END :
		pop->highlight(pop->list.count-1, true);
		break;

	case VK_NEXT :
	{
		int c = pop->current;
		int lv = pop->lastVisible();
		if ( lv < pop->list.count-1) 
			lv--;
		if (c < lv)
			c = lv;
		else
			if ((c += pop->numVisible-1) > pop->list.count-1)
				c = pop->list.count-1;
		pop->highlight(c, true);
		break;
	}

	case VK_PRIOR :
	{
		int c = pop->current;
		int fv = pop->firstVisible();
		if (fv > 0) 
			fv++;
		if (c > fv)
			c = fv;
		else
			if ((c -= pop->numVisible-1) < 0)
				c = 0;
		pop->highlight(c, true);
		break;
	}

	case VK_RETURN :
		pop->execute();
		break;

	case VK_ESCAPE :
		popup.destroyWindow();
		break;
	}
	break;

	case WM_NCDESTROY:
	case WM_CLOSE:
	case WM_DESTROY:
	case WM_QUIT:
	case WM_ENDSESSION:
	case WM_QUERYENDSESSION:
	break;
	default:
		return ::DefWindowProc(hwnd, message, wParam, lParam);
	}
	return ::DefWindowProc(hwnd, message, wParam, lParam);
}

//cleanup, not much to do
int quitModule(HINSTANCE dllInst) {
	popup.destroyWindow();
	::DestroyWindow(popup.hwnd);
    ::DeleteObject(popupFont);
	::DeleteObject(hpenWhite);
	::DeleteObject(hpenBlack);
	::DeleteObject(hpenGray);
	::DeleteObject(hbrBlack);
	::DeleteObject(hbrGray);
	if (showPopupIcons) {
		DeleteObject(popupDefaultIcon);
	}
    UnregisterClass(szAppName,dllInst); // unregister window class
    return 1;
}

//now defunct entry point, will take it out if somone complains
int initModule(HWND ParentWnd, HINSTANCE dllInst, wharfDataType* wd)
{
    return initModuleEx (ParentWnd, dllInst, wd->lsPath);
}

//DLL entry point
int initModuleEx(HWND ParentWnd, HINSTANCE dllInst, LPCSTR szPath)
{
	WNDCLASSEX wc;

	_RPT0(_CRT_WARN, "Starting Popup module");
    appInstance = dllInst;
    parentWin = ParentWnd;
    
    memset(&wc,0,sizeof(wc));
	wc.cbSize = sizeof(wc);
    wc.style = 0;
    wc.lpfnWndProc = (WNDPROC)wndProc;// our window procedure
    //wc.hbrBackground = (HBRUSH) GetStockObject(LTGRAY_BRUSH);
    wc.hInstance = appInstance;			
	wc.cbWndExtra = 8; // signature + Popup object address
    wc.lpszClassName = szAppName;// our window class name
    if (!RegisterClassEx(&wc)) {
		_RPT1(_CRT_ERROR, "Failed to register POPUP window class; error '%d'\n", GetLastError());
    	return 1;
    }
	_RPT0(_CRT_WARN, "Registered Popup class");

	popup.loadValues();
	_RPT0(_CRT_WARN, "Loaded values");
	popup.createWindow();
	_RPT0(_CRT_WARN, "Created main popup");
	if (!popup.hwnd) {
		_RPT0(_CRT_ERROR, "Failed to create main POPUP window");
		return 1;
	}
/*	if (transBlt && !noPopupPix) {
		SetWindowRgn(popup.hwnd, popup.region, FALSE);
	} */

    //create the message id's that popup.dll will use
    int Msgs[10];
    Msgs[0] = LM_LPOPUP;
    Msgs[1] = LM_HIDEPOPUP;
    Msgs[2] = LM_GETREVID;
	Msgs[3] = 0;
    SendMessage (ParentWnd, LM_REGISTERMESSAGE, (WPARAM)popup.hwnd, (LPARAM)Msgs);
    return 0;
}
