#ifndef __POPUP_H
#define __POPUP_H

#include "../litestep/wharfdata.h"
#include "../lsapi/lsapi.h"

const MAX_TEXT = 256;
const MENU_TIMER = 701;
const SCROLL_TIMER = 702;
const SCROLL_DELAY = 10;

//pre-decl 
class Popup;

//simple bitmap class
class BitMP {
public:
	HBITMAP bitmap;
	HRGN region;
	char name[MAX_TEXT];
	int x;
	int y;
	COLORREF backColor;
	COLORREF foreColor;

	BitMP();
	~BitMP() {destroy();}
protected:
	void destroy();
};

class menuItem {
public:
    enum itemType {mi_basic, mi_folder, mi_static, mi_dynamic, mi_task};
	Popup *owner;
	char name[MAX_TEXT];
	char command[MAX_TEXT];
	char params[MAX_TEXT];
	Popup *sub;
	HICON hIcon;
	itemType type;

	menuItem();
	menuItem(LPSTR lpszText, LPSTR lpszCommand, LPSTR lpszParams = NULL, Popup* pop = NULL);
	~menuItem();
protected:
	void init();
};

// the list class
template<class T> class list {
public:
    int  count;
    T   *data;

public:
	list(int alloc = 16, int inc = 8);
	virtual ~list();
	T& operator [] (int i);
	const T& operator [] (int i) const;
	int  add(const T& t);
	void copy( const list<T>& l);
	void clear();
	void sort(int (__cdecl *compare )(const void *elem1, const void *elem2));
	bool validIndex(int indx) const;

protected:
	void reallocate(int num = 0);

protected:
    int  increment;
    int  allocated;
};

// the menu item list
class itemList : public list<menuItem> {
public:
	int add(LPSTR lpszText, LPSTR lpszCommand, LPSTR lpszParams = NULL, Popup *lp = NULL);
};

//the popup list container
class Popup 
{ 
public:
	char title[MAX_TEXT];
	int current;
	int first;
	int last;
	int numVisible;
	HWND hwnd;
	itemList list;
	static BOOL init;
public:
	Popup(Popup *parent = NULL);
	~Popup();
    friend LRESULT CALLBACK wndProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam);

    menuItem*  safeItem(int indx) const;
    int  firstVisible();
    int  lastVisible();
    void firstVisible(int n);
    void lastVisible(int n);
	void scroll(bool downd);
	bool createWindow();
	void destroyWindow();
	void display(int x, int y);
	void display(POINT pt) {display(pt.x, pt.y);}
	void paint();
	void paintBMP(HDC dest, HDC src, BitMP bmp, RECT r);
	void drawItem(const menuItem& mi, HDC hdc, HDC mem, BitMP bmp, RECT r);
	bool itemRect(int indx, RECT *lpRect) const;
	int calcSize(SIZE& s) const;
	int  itemAt(POINT pt) const;
	bool highlight(int indx, bool snapCursor = false);
	int  hittest(POINT pt, int *pindx = NULL) const; // HTCAPTION/HTCLIENT/HTNOWHERE
	bool execute(int indx = -1); // -1 means execute 'current'
	enum {
		gi_clsid = 0x01,   //lookup among predefined CLSIDs
		gi_shell = 0x02    //ask shell to provide icon info
	};
	void buildTaskList();

	void loadValues();
};

struct timer {
	int id;
	bool timerSet;
	Popup  *tmPop;
	int tmNdx;   // item in this popup

	timer(int i) {id=i, timerSet=false, tmPop=NULL, tmNdx=-1;}
	void setTimer(Popup *p, int indx, int interval = 0);
	void killTimer();
	bool sameItem(Popup *p, int indx) const;
	void resetIndex() {tmNdx = -1;}
};

#ifdef	__cplusplus
extern "C" {
#endif	/* __cplusplus */
__declspec( dllexport ) int initModule(HWND ParentWnd, HINSTANCE dllInst, wharfDataType* wd);
__declspec( dllexport ) int initModuleEx(HWND parent, HINSTANCE dll, LPCSTR szPath);
__declspec( dllexport ) int quitModule(HINSTANCE dllInst);

#ifdef	__cplusplus
};
#endif	/* __cplusplus */

#endif  /* __LSAPI_H */
